package com.fr.plugin.sxka.sso;

import com.fr.base.PropertiesUtils;
import com.fr.data.NetworkHelper;
import com.fr.decision.authority.AuthorityContext;
import com.fr.decision.authority.data.CustomRole;
import com.fr.decision.authority.data.User;
import com.fr.decision.fun.impl.AbstractGlobalRequestFilterProvider;
import com.fr.decision.mobile.terminal.TerminalHandler;
import com.fr.decision.webservice.bean.authentication.OriginUrlResponseBean;
import com.fr.decision.webservice.bean.user.UserBean;
import com.fr.decision.webservice.exception.user.UserNotExistException;
import com.fr.decision.webservice.utils.DecisionServiceConstants;
import com.fr.decision.webservice.utils.DecisionStatusService;
import com.fr.decision.webservice.utils.WebServiceUtils;
import com.fr.decision.webservice.v10.login.LoginService;
import com.fr.decision.webservice.v10.login.TokenResource;
import com.fr.decision.webservice.v10.user.UserService;
import com.fr.general.http.HttpRequest;
import com.fr.general.http.HttpToolbox;
import com.fr.json.JSONObject;
import com.fr.json.revise.EmbedJson;
import com.fr.locale.InterProviderFactory;
import com.fr.log.FineLoggerFactory;
import com.fr.plugin.sxka.sso.dao.UserInfoDao;
import com.fr.plugin.sxka.sso.entity.UserInfoEntity;
import com.fr.plugin.transform.FunctionRecorder;
import com.fr.security.encryption.transmission.TransmissionEncryptors;
import com.fr.stable.StringUtils;
import com.fr.stable.db.action.DBAction;
import com.fr.stable.db.dao.DAOContext;
import com.fr.stable.query.QueryFactory;
import com.fr.stable.query.restriction.RestrictionFactory;
import com.fr.stable.web.Device;
import com.fr.third.fasterxml.jackson.databind.DeserializationFeature;
import com.fr.third.fasterxml.jackson.databind.ObjectMapper;
import com.fr.third.springframework.http.HttpHeaders;
import com.fr.third.springframework.http.MediaType;
import com.fr.third.springframework.util.DigestUtils;
import com.fr.web.controller.decision.api.auth.LoginResource;
import com.fr.web.utils.WebUtils;

import javax.servlet.FilterChain;
import javax.servlet.ServletException;
import javax.servlet.http.Cookie;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.io.PrintWriter;
import java.net.URLEncoder;
import java.nio.charset.StandardCharsets;
import java.util.*;

/**
 * @Author fr,open
 * @Date 2021/5/12
 * @Description
 **/
@FunctionRecorder
public class SsoFilter extends AbstractGlobalRequestFilterProvider {

    private static String[] notFilter = {
            "/login",
            "/system", "/decision/file",
            "/decision/resources", "/remote","/view/report"
    };
    private static HashMap<String, String> ROLE_MAP;

    static {
        ROLE_MAP = new HashMap<>();
        ROLE_MAP.put("1", "企业管理员");
        ROLE_MAP.put("2", "个人用户");
        ROLE_MAP.put("3", "企业操作员");
        ROLE_MAP.put("4", "政府端用户");
    }

    @Override
    public String filterName() {
        return "sxka";
    }

    @Override
    public String[] urlPatterns() {
        return new String[]{"/*"};
    }

    @Override
    public void doFilter(HttpServletRequest req, HttpServletResponse res, FilterChain filterChain) {
        if(req.getRequestURI().endsWith("/test/error")){
            setError(res);
            return;
        }
        if (req.getRequestURI().endsWith("/admin/login")) {
            try {
                String page = new LoginResource().page(req, res);
                if (StringUtils.isBlank(page)) {
                    return;
                }
                PrintWriter writer = WebUtils.createPrintWriter(res);
                writer.println(page);
                writer.flush();
                writer.close();
            } catch (Exception e) {
                FineLoggerFactory.getLogger().error(e.getMessage(), e);
            }
            return;
        }
        //登录页面跳转地址拦截
        String origin = WebUtils.getHTTPRequestParameter(req, "origin");
        if (req.getRequestURI().endsWith("decision/login") && "get".equalsIgnoreCase(req.getMethod())) {
            String state = this.getUserStateFromCookie(req, res);

            try {
                if (StringUtils.isNotBlank(origin)) {
                    OriginUrlResponseBean path = DecisionStatusService.originUrlStatusService().get(origin);
                    cacheParam(res, path.getOriginUrl());
                }
                res.sendRedirect(buildAuthorizeUrl(state));
            } catch (Exception e) {
                FineLoggerFactory.getLogger().error(e.getMessage(), e);
            }
        }
        if (isLogin(req) || isAccept(req)) {
            filter(req, res, filterChain);
            return;
        }
        String state = this.getUserStateFromCookie(req, res);
        String code = req.getParameter("code");
        if (StringUtils.isNotBlank(code)) {
            JSONObject data = acquireAccessToken(code);
            FineLoggerFactory.getLogger().info("get access_token is {}", data);
            // 计算过期时间
            int expires = Integer.valueOf(data.get("expires_in").toString());
            Date expiresTime = new Date(System.currentTimeMillis() + (expires * 1000));
            data.put("expires_at", expiresTime);
            // 保存关系映射
            putState(state, data);
        }
        JSONObject tokenItem = null;
        try {
            tokenItem = DecisionStatusService.loginStatusService().get(state);
        } catch (Exception e) {
            FineLoggerFactory.getLogger().error(e.getMessage(),e);
        }
        if (tokenItem == null) {
            try {
                String jump = req.getRequestURL() + (req.getQueryString() == null ? StringUtils.EMPTY : "?" + req.getQueryString());
                cacheParam(res, jump);
                res.sendRedirect(buildAuthorizeUrl(state));
                return;
            } catch (Exception e) {
                FineLoggerFactory.getLogger().error(e.getMessage(), e);
            }
        }
        // 用检查 access token 是否过期
        Date tokenExpiresAt = (Date) tokenItem.get("expires_at");
        Date now = new Date();
        if (tokenExpiresAt.before(now)) {
            // 使用 refresh token 刷新 access token
            String refreshToken = tokenItem.get("refresh_token").toString();
            tokenItem = this.refreshToken(refreshToken);
            putState(state, tokenItem);
        }
        String accessToken = tokenItem.get("access_token").toString();
        JSONObject userProfile = this.acquireUserInfo(state, accessToken);
        try {
            if (!existUser(userProfile.getString("id"))) {
                FineLoggerFactory.getLogger().info("user {} not exist ", userProfile.getString("id"));
                UserBean userBean = new UserBean();
                userBean.setUsername(userProfile.getString("id"));
                userBean.setRealName(userProfile.getString("op_name"));
                userBean.setEnable(true);
                userBean.setPassword(TransmissionEncryptors.getInstance().encrypt("!234Qwer"));
                UserService.getInstance().addUser(userBean);
            }
            UserInfoEntity infoEntity = recordUserInfo(userProfile);
            if(checkAuth(infoEntity, req, res)){
                return;
            }
        } catch (Exception e) {
            FineLoggerFactory.getLogger().error(e.getMessage(), e);
        }
        filter(req, res, filterChain);
    }

    private void putState(String state, JSONObject data) {
        try {
            DecisionStatusService.loginStatusService().put(state, data);
        } catch (Exception e) {
            FineLoggerFactory.getLogger().error(e.getMessage(),e);
        }
    }

    private void cacheParam(HttpServletResponse res, String jump) {
        String id = UUID.randomUUID().toString();
        try {
            DecisionStatusService.originUrlStatusService().put(id, new OriginUrlResponseBean(jump));//添加重定向地址
        } catch (Exception e) {
            FineLoggerFactory.getLogger().error(e.getMessage(), e);
        }
        Cookie cookie = new Cookie("ORIGIN_URL", id);
        cookie.setPath("/");
        cookie.setMaxAge(60);
        cookie.setHttpOnly(true);
        res.addCookie(cookie);
    }

    private boolean checkAuth(UserInfoEntity entity, HttpServletRequest req, HttpServletResponse res) throws Exception {
        if (StringUtils.isNotBlank(entity.getStatus()) && entity.getStatus().equalsIgnoreCase("true")) {
            if (entity.getAccount_type().equals("1") || entity.getAccount_type().equals("3") || entity.getAccount_type().equals("4")) {
                UserBean userAccount = UserService.getInstance().getUserAccount(entity.getId());
                String roleName = ROLE_MAP.containsKey(entity.getAccount_type()) ? ROLE_MAP.get(entity.getAccount_type()) : "其他";
                CustomRole role = AuthorityContext.getInstance().getCustomRoleController().findOne(QueryFactory.create().addRestriction(RestrictionFactory.eq("name", roleName)));
                if (role != null) {
                    List<CustomRole> oldRoleList = AuthorityContext.getInstance().getCustomRoleController().findByUser(userAccount.getId(), QueryFactory.create());
                    oldRoleList.forEach(e->{
                        try {
                            AuthorityContext.getInstance().getUserController().removeUserFromCustomRole(userAccount.getId(), e.getId());
                        } catch (Exception exception) {
                            FineLoggerFactory.getLogger().error(exception.getMessage(),exception);
                        }
                    });
                    AuthorityContext.getInstance().getUserController().addUserToCustomRole(userAccount.getId(),role.getId());
                }
            }
            loginFromToken(req, res, entity.getId());
            String jump = getOrigin(req);
            FineLoggerFactory.getLogger().info("get Origin path is {}", jump);
            if (StringUtils.isNotBlank(jump)) {
                res.sendRedirect(jump);
                return true;
            }
            return false;
        }
        if(entity.getAccount_type().equals("1") || entity.getAccount_type().equals("3") || entity.getAccount_type().equals("4")){
            setError(res);
            return true;
        }
        String url = PropertiesUtils.getProperties("conf").getProperty("write_report");
        String param = "username="+ URLEncoder.encode(entity.getOp_name(),"UTF-8")+"&id="+entity.getId();
        url = url.indexOf("?") != -1?url+"&"+param:url+"?"+param;
        res.sendRedirect(url);
        return true;
    }



    private String getOrigin(HttpServletRequest req) throws Exception {
        Cookie[] cookies = req.getCookies();
        if (cookies == null) {
            return null;
        }
        for (int i = 0; i < cookies.length; i++) {
            if ("ORIGIN_URL".equals(cookies[i].getName())) {
                String id = cookies[i].getValue();
                OriginUrlResponseBean path = DecisionStatusService.originUrlStatusService().get(id);
                if (path != null) {
                    DecisionStatusService.originUrlStatusService().delete(id);
                    return path.getOriginUrl();
                }
            }
        }
        return StringUtils.EMPTY;
    }

    private UserInfoEntity recordUserInfo(JSONObject user) {
        ObjectMapper mapper = EmbedJson.MAPPER.configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);
        UserInfoEntity entity = mapper.convertValue(user, UserInfoEntity.class);
        try {
            return RoleDBAccessProvider.getDbAccessor().runDMLAction(new DBAction<UserInfoEntity>() {
                @Override
                public UserInfoEntity run(DAOContext daoContext) throws Exception {
                    UserInfoDao dao = daoContext.getDAO(UserInfoDao.class);
                    UserInfoEntity byId = dao.getById(entity.getId());
                    if (byId != null) {
                        entity.setStatus(byId.getStatus());
                        entity.setAccount_type(byId.getAccount_type());
                    } else {
                        entity.setStatus("false");
                    }
                    dao.addOrUpdate(entity);
                    return entity;
                }
            });
        } catch (Exception e) {
            FineLoggerFactory.getLogger().error(e.getMessage(), e);
        }
        return entity;

    }

    private boolean existUser(String username) {
        User user = null;
        try {
            user = UserService.getInstance().getUserByUserName(username);
        } catch (Exception e) {
            FineLoggerFactory.getLogger().error(e.getMessage(), e);
        }
        return user != null;
    }

    private boolean loginFromToken(HttpServletRequest req, HttpServletResponse res, String username) throws Exception {
        try {
            if (StringUtils.isNotEmpty(username)) {
                FineLoggerFactory.getLogger().info("current username:" + username);
                User user = UserService.getInstance().getUserByUserName(username);
                FineLoggerFactory.getLogger().info("get user:" + user);
                if (user == null) {
                    throw new UserNotExistException();
                }
                String token = LoginService.getInstance().login(req, res, username);
                FineLoggerFactory.getLogger().info("get login token:" + token);
                req.setAttribute(DecisionServiceConstants.FINE_AUTH_TOKEN_NAME, token);
                FineLoggerFactory.getLogger().info("username:" + username + "login success");
                return true;
            } else {
                FineLoggerFactory.getLogger().warn("username is null!");
                return false;
            }
        } catch (Exception e) {
            FineLoggerFactory.getLogger().error(e.getMessage(), e);
        }
        return false;
    }


    private JSONObject acquireUserInfo(String state, String accessToken) {
        String url = PropertiesUtils.getProperties("conf").getProperty("user_url");

        Map<String, String> params = new HashMap<>();
        params.put("client_id", PropertiesUtils.getProperties("conf").getProperty("client_id"));
        params.put("access_token", accessToken);
        HttpHeaders headers = new HttpHeaders();
        headers.setContentType(MediaType.APPLICATION_FORM_URLENCODED);
        try {
            String res = HttpToolbox.executeAndParse(HttpRequest.custom().url(url)
                    .headers(headers.toSingleValueMap())
                    .post(params)
                    .build());
            FineLoggerFactory.getLogger().info("get user name url is 【{}】,params is，{} res is {}", url, params, res);
            JSONObject object = new JSONObject(res);
            return object.getJSONObject("data");
        } catch (IOException e) {
            FineLoggerFactory.getLogger().error(e.getMessage(), e);
        }
        return JSONObject.EMPTY;
    }

    private JSONObject refreshToken(String refreshToken) {
        String url = PropertiesUtils.getProperties("conf").getProperty("token_url");

        Map<String, String> params = new HashMap<>();
        params.put("grant_type", "refresh_token");
        params.put("refresh_token", refreshToken);
        params.put("client_id", PropertiesUtils.getProperties("conf").getProperty("client_id"));
        params.put("client_sign", this.makeSign(refreshToken));
        HttpHeaders headers = new HttpHeaders();
        headers.setContentType(MediaType.APPLICATION_FORM_URLENCODED);
        try {
            String res = HttpToolbox.executeAndParse(HttpRequest.custom().url(url)
                    .headers(headers.toSingleValueMap())
                    .post(params)
                    .build());
            FineLoggerFactory.getLogger().info("get refresh token url is 【{}】,params is，{} res is {}", url, params, res);
            JSONObject object = new JSONObject(res);
            return object.getJSONObject("data");
        } catch (IOException e) {
            FineLoggerFactory.getLogger().error(e.getMessage(), e);
        }
        return JSONObject.EMPTY;
    }

    private JSONObject acquireAccessToken(String code) {
        String url = PropertiesUtils.getProperties("conf").getProperty("token_url");
        Map<String, String> params = new HashMap<>();
        params.put("grant_type", "authorization_code");
        params.put("code", code);
        params.put("client_id", PropertiesUtils.getProperties("conf").getProperty("client_id"));
        params.put("client_sign", this.makeSign(code));
        HttpHeaders headers = new HttpHeaders();
        headers.setContentType(MediaType.APPLICATION_FORM_URLENCODED);
        try {
            String res = HttpToolbox.executeAndParse(HttpRequest.custom().url(url)
                    .headers(headers.toSingleValueMap())
                    .post(params)
                    .build());
            FineLoggerFactory.getLogger().info("get access token url is 【{}】,params is，{} res is {}", url, params, res);
            JSONObject object = new JSONObject(res);
            return object.getJSONObject("data");
        } catch (IOException e) {
            FineLoggerFactory.getLogger().error(e.getMessage(), e);
        }
        return JSONObject.EMPTY;
    }

    /**
     * 生成签名值= md5Hex("<client_id>$<data>$<client_secret>", UTF8)
     *
     * @param data
     * @return
     */
    private String makeSign(String data) {
        String value = String.join("$", new String[]{
                PropertiesUtils.getProperties("conf").getProperty("client_id"),
                data,
                PropertiesUtils.getProperties("conf").getProperty("client_secret")
        });
        return DigestUtils.md5DigestAsHex(value.getBytes(StandardCharsets.UTF_8));
    }

    private String buildAuthorizeUrl(String state) {
        String format = "%s?client_id=%s&response_type=code&scope=A1,C1,C2&state=%s&redirect_uri=%s";
        return String.format(format, PropertiesUtils.getProperties("conf").getProperty("auth_url"), PropertiesUtils.getProperties("conf").getProperty("client_id"), state, PropertiesUtils.getProperties("conf").getProperty("redirect_uri"));
    }

    private void setUserStateCookie(HttpServletResponse response, String state) {
        Cookie cookie = new Cookie("uid_state", state);
        cookie.setMaxAge(300);
        // 正式环境采用HTTPS协议，需要取消下方注释，确保仅在HTTPS下可用
//        cookie.setSecure(true);
        cookie.setPath("/");
        response.addCookie(cookie);
    }

    private String getUserStateFromCookie(HttpServletRequest request, HttpServletResponse res) {
        if (request.getCookies() == null) return null;

        String state = StringUtils.EMPTY;
        for (Cookie cookie : request.getCookies()) {
            if ("uid_state".equals(cookie.getName())) {
                state = cookie.getValue();
            }
        }
        if (StringUtils.isBlank(state)) {
            state = UUID.randomUUID().toString();
            this.setUserStateCookie(res, state);
        }
        return state;
    }

    private boolean isAccept(HttpServletRequest req) {
        for (int i = 0; i < notFilter.length; i++) {
            if (req.getRequestURI().contains(notFilter[i])) {
                return true;
            }
        }
        if (req.getRequestURI().endsWith("/.css") || req.getRequestURI().endsWith("/.css")) {
            return true;
        }
        return false;
    }

    private boolean isLogin(HttpServletRequest request) {
        String oldToken = TokenResource.COOKIE.getToken(request);
        return oldToken != null && checkTokenValid(request, (String) oldToken);
    }

    private boolean checkTokenValid(HttpServletRequest req, String token) {
        try {
            Device device = NetworkHelper.getDevice(req);
            LoginService.getInstance().loginStatusValid(token, TerminalHandler.getTerminal(req, device));
            return true;
        } catch (Exception ignore) {
        }
        return false;
    }

    private void setError(HttpServletResponse res) {
        try {
            PrintWriter printWriter = WebUtils.createPrintWriter(res);
            Map<String, Object> map = new HashMap<>();
            map.put("result", InterProviderFactory.getProvider().getLocText("Fine-Engine_Error_Page_Result"));
            map.put("reason", "当前用户已冻结，请联系系统管理员激活");
            map.put("solution", InterProviderFactory.getProvider().getLocText("Fine-Engine_Please_Contact_Platform_Admin"));
            String page = WebServiceUtils.parseWebPageResourceSafe("com/fr/web/controller/decision/entrance/resources/unavailable.html", map);
            printWriter.write(page);
            printWriter.flush();
            printWriter.close();
        } catch (Exception e) {
            FineLoggerFactory.getLogger().error(e.getMessage(), e);
        }
    }

    private void filter(HttpServletRequest req, HttpServletResponse res, FilterChain filterChain) {
        try {
            filterChain.doFilter(req, res);
        } catch (IOException e) {
            FineLoggerFactory.getLogger().error(e.getMessage(), e);
        } catch (ServletException e) {
            FineLoggerFactory.getLogger().error(e.getMessage(), e);
        }
    }
}
